use crate::{
    color::Rgba,
    geometry::{Point, Point2, Ray, Transform, Transform2, Vector, point2, ray, vec3},
    sampling::PixelSample,
};

#[derive(Debug, Clone)]
pub struct Camera {
    /// camera to world transform
    transform: Transform,
}

impl Camera {
    pub fn new(eye: Point, center: Point, up: Vector, fov: f32) -> Self {
        let f = (center - eye).length();
        let tan = (fov / 2.0).to_radians().tan();
        let transform =
            Transform::look_at(eye, center, up).inverse() * Transform::scale(tan * f, tan * f, f);

        Self { transform }
    }

    pub fn ray(&self, p_film: Point2) -> Ray {
        ray(
            self.transform * Point::ORIGIN,
            self.transform * vec3(p_film.x, p_film.y, 1.0),
        )
    }
}

pub struct Film {
    width: u32,
    height: u32,
    transform: Transform2,
    data: Box<[Rgba]>,
}

impl Film {
    pub fn new(width: u32, height: u32) -> Self {
        let size = width as usize * height as usize;

        let data = vec![Rgba::BLACK; size].into_boxed_slice();

        let w = width as f32;
        let h = height as f32;
        let ar = h / w;

        let transform = Transform2::translate(-1.0, ar) * Transform2::scale(2.0 / w, -2.0 / w);

        Self {
            width,
            height,
            transform,
            data,
        }
    }

    pub fn add_sample(&mut self, x: u32, y: u32, color: Rgba, weight: f32) {
        let i = y as usize * self.width as usize + x as usize;
        self.data[i] += color * weight;
    }

    pub fn get_camera_sample(&self, x: u32, y: u32, sampler: &mut PixelSample) -> Point2 {
        let p = point2(x as f32, y as f32) + sampler.get_2d().into();
        self.transform * p
    }

    pub fn into_data(self) -> Box<[Rgba]> {
        self.data
    }
}
