use std::fmt;

use enum_dispatch::enum_dispatch;

use super::{Point, Ray, Vector};

#[derive(Debug)]
pub struct Hit {
    pub point: Point,
    pub normal: Vector,
    pub t: f32,
}

#[enum_dispatch]
pub trait Hittable {
    fn intersect(&self, ray: &Ray, t_max: f32) -> Option<Hit>;
}

#[derive(Debug, Clone)]
pub struct Sphere {
    center: Point,
    radius: f32,
}

impl Hittable for Sphere {
    fn intersect(&self, ray: &Ray, t_max: f32) -> Option<Hit> {
        let oc = self.center - ray.origin;
        let a = ray.direction.length_squared();
        let h = ray.direction.dot(oc);
        let c = oc.length_squared() - self.radius * self.radius;
        let d = h * h - a * c;

        (d > 0.0)
            .then(|| (h - d.sqrt()) / a)
            .filter(|t| *t < t_max && *t > 0.0)
            .map(|t| {
                let point = ray.at(t);
                let normal = point - self.center;
                Hit { point, normal, t }
            })
    }
}

#[derive(Debug, Clone)]
pub struct Plane {
    normal: Vector,
    d: f32,
}

impl Hittable for Plane {
    fn intersect(&self, ray: &Ray, t_max: f32) -> Option<Hit> {
        let n_dot_r = self.normal.dot(ray.direction);
        (n_dot_r.abs() > f32::EPSILON)
            .then(|| {
                let n_dot_o = self.normal.dot(ray.origin.into());
                -(n_dot_o - self.d) / n_dot_r
            })
            .filter(|t| *t < t_max && *t > 0.0)
            .map(|t| Hit {
                point: ray.at(t),
                normal: self.normal,
                t,
            })
    }
}

#[enum_dispatch(Hittable)]
#[derive(Clone)]
pub enum Shape {
    Sphere,
    Plane,
}

impl fmt::Debug for Shape {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Sphere(s) => s.fmt(f),
            Self::Plane(p) => p.fmt(f),
        }
    }
}

impl Shape {
    pub fn sphere(center: Point, radius: f32) -> Shape {
        Shape::from(Sphere { center, radius })
    }

    pub fn plane(normal: Vector, d: f32) -> Shape {
        Shape::from(Plane { normal, d })
    }
}
