use std::{fmt, ops::Mul};

use super::{Point, Ray, Vector};

#[derive(Copy, Clone)]
#[repr(transparent)]
pub struct Transform(glam::Affine3A);

impl fmt::Debug for Transform {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct(stringify!(Transform))
            .field(stringify!(matrix3), &self.0.matrix3)
            .field(stringify!(translation), &self.0.translation)
            .finish()
    }
}

impl Transform {
    pub fn scale(x: f32, y: f32, z: f32) -> Self {
        Self(glam::Affine3A::from_scale(glam::vec3(x, y, z)))
    }

    pub fn inverse(self) -> Self {
        Self(self.0.inverse())
    }

    pub fn look_at(eye: Point, center: Point, up: Vector) -> Self {
        Self(glam::Affine3A::look_at_lh(
            eye.into(),
            center.into(),
            up.into(),
        ))
    }
}

impl Mul for Transform {
    type Output = Transform;

    fn mul(self, rhs: Self) -> Self::Output {
        Self(self.0 * rhs.0)
    }
}

impl Mul<Vector> for Transform {
    type Output = Vector;

    fn mul(self, rhs: Vector) -> Self::Output {
        Vector(self.0.transform_vector3a(rhs.0))
    }
}

impl Mul<Point> for Transform {
    type Output = Point;

    fn mul(self, rhs: Point) -> Self::Output {
        Point(self.0.transform_point3a(rhs.0))
    }
}

impl Mul<Ray> for Transform {
    type Output = Ray;

    fn mul(self, rhs: Ray) -> Self::Output {
        Ray {
            origin: self * rhs.origin,
            direction: self * rhs.direction,
        }
    }
}
