use core::f32;

use geometry::{Hit, Hittable, Ray, point2, vec3};
use range2d::Range2D;

#[cfg(feature = "bytemuck")]
pub mod impl_bytemuck;

pub mod color;

mod camera;
mod geometry;
mod sampling;

pub use camera::Camera;
pub use geometry::{Point, Shape, Transform, Vector};

use camera::Film;
use color::{Rgba, rgba};
use sampling::Sampler;

pub struct Scene {
    camera: Camera,
    objects: Vec<Shape>,
}

impl Scene {
    pub fn new(camera: Camera, objects: Vec<Shape>) -> Self {
        Self { camera, objects }
    }

    pub fn render(&self, width: u32, height: u32, samples: u32, seed: u32) -> Box<[Rgba]> {
        let mut sampler = Sampler::new(width, height, samples, seed);
        let mut film = Film::new(width, height);

        for (y, x) in Range2D::new(0..height, 0..width) {
            let mut rgb = Vector::ZERO;

            for mut pixel_sample in sampler.pixel(x, y) {
                let p_film = film.get_camera_sample(x, y, &mut pixel_sample);
                let ray = self.camera.ray(p_film);

                if let Some(h) = self.intersect(ray) {
                    rgb += h.normal.normalize() * 0.5 + vec3(0.5, 0.5, 0.5);
                }
            }

            rgb /= samples as f32;

            film.add_sample(x, y, rgba(rgb.x, rgb.y, rgb.z, 0.0), 1.0);
        }

        film.into_data()
    }

    fn intersect(&self, r: Ray) -> Option<Hit> {
        let mut hit: Option<Hit> = None;

        for obj in &self.objects {
            let t_max = hit.as_ref().map(|h| h.t).unwrap_or(f32::INFINITY);
            if let Some(h) = obj.intersect(&r, t_max) {
                hit = Some(h)
            }
        }

        hit
    }
}
