use bytemuck::allocation::cast_slice_box;
use mechthild_core::{Camera, Scene, Shape, color::Rgba};
use pyo3::prelude::*;

#[pyclass(name = "Camera")]
#[derive(Clone)]
struct PyCamera(Camera);

#[pymethods]
impl PyCamera {
    #[new]
    pub fn new(
        eye: (f32, f32, f32),
        center: (f32, f32, f32),
        up: (f32, f32, f32),
        fov: f32,
    ) -> Self {
        Self(Camera::new(eye.into(), center.into(), up.into(), fov))
    }

    pub fn __str__(&self) -> PyResult<String> {
        Ok(format!("{:#?}", self.0))
    }
}

#[pyclass(name = "Shape")]
#[derive(Clone)]
struct PyShape(Shape);

#[pymethods]
impl PyShape {
    #[staticmethod]
    pub fn sphere(center: (f32, f32, f32), radius: f32) -> Self {
        Self(Shape::sphere(center.into(), radius))
    }

    #[staticmethod]
    pub fn plane(normal: (f32, f32, f32), d: f32) -> Self {
        Self(Shape::plane(normal.into(), d))
    }

    pub fn __str__(&self) -> PyResult<String> {
        Ok(format!("{:?}", self.0))
    }
}

#[pyclass(name = "Scene")]
struct PyScene(Scene);

#[pymethods]
impl PyScene {
    #[new]
    pub fn new(camera: &PyCamera, contents: Vec<PyShape>) -> Self {
        let contents = contents.into_iter().map(|s| s.0).collect();
        Self(Scene::new(camera.0.clone(), contents))
    }

    #[pyo3(signature = (width, height, samples, seed = 0))]
    pub fn render(&self, width: u32, height: u32, samples: u32, seed: u32) -> Vec<f32> {
        let result = self.0.render(width, height, samples, seed);

        cast_slice_box::<Rgba, f32>(result).into_vec()
    }
}

#[pymodule]
fn mechthild(m: &Bound<'_, PyModule>) -> PyResult<()> {
    m.add_class::<PyCamera>()?;
    m.add_class::<PyShape>()?;
    m.add_class::<PyScene>()?;
    Ok(())
}
